/*
 * Copyright (c) 1998-2000 Apple Computer, Inc. All rights reserved.
 *
 * @APPLE_LICENSE_HEADER_START@
 * 
 * The contents of this file constitute Original Code as defined in and
 * are subject to the Apple Public Source License Version 1.1 (the
 * "License").  You may not use this file except in compliance with the
 * License.  Please obtain a copy of the License at
 * http://www.apple.com/publicsource and read it before using this file.
 * 
 * This Original Code and all software distributed under the License are
 * distributed on an "AS IS" basis, WITHOUT WARRANTY OF ANY KIND, EITHER
 * EXPRESS OR IMPLIED, AND APPLE HEREBY DISCLAIMS ALL SUCH WARRANTIES,
 * INCLUDING WITHOUT LIMITATION, ANY WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE OR NON-INFRINGEMENT.  Please see the
 * License for the specific language governing rights and limitations
 * under the License.
 * 
 * @APPLE_LICENSE_HEADER_END@
 */

#include <IOKit/system.h>

#include <libkern/c++/OSContainers.h>
#include <libkern/OSByteOrder.h>
#include <IOKit/IODeviceMemory.h>
#include <IOKit/IODeviceTreeSupport.h>
#include <IOKit/IOLib.h>
#include <IOKit/IOMapper.h>
#include <IOKit/acpi/IOACPIPlatformDevice.h>

#include "AGPGart.h"
#include "pcireg.h"

/****************************************************************************/
//	AGP GART Driver port for Mac OSX x86
/****************************************************************************/
#define AGP_ATTBASE 0x18

//#define INGART(offset) OSReadLittleInt32((agpio_base), offset)
//#define OUTGART(offset,val) OSWriteLittleInt32((agpio_base), offset, val)


//Risc
#define super IOPCIBridge
//#define super AppleMacRiscPCI

OSDefineMetaClassAndStructors(AGPGart, IOPCIBridge); //AppleMacRiscPCI)

//bool isIntel;
//static IOSimpleLock *  	 AGPBridgeLock;
//----------------------------------------------------------------------------

IOService * 
AGPGart::probe( IOService * provider, SInt32 * score )
{
	OSData *data;
	UInt16 deviceID;
	const char* str;
//	isIntel = false;
	
	

	data = OSDynamicCast(OSData, provider->getProperty(kDevice_IDProperty));	
	deviceID = *((UInt16 *) data->getBytesNoCopy());
	
	switch(GetVendor(provider))
	{
		case ATI:
			bridgeDev = new AGPATI(provider);
			if((str = bridgeDev->Match(deviceID)) == NULL)
			{
				bridgeDev->release();
				bridgeDev = new AGPAMD64(provider);
				str = bridgeDev->Match(deviceID);
			}
			break;
		case INTEL:
			bridgeDev = new AGPINTEL(provider);
			str = bridgeDev->Match(deviceID);
	//		isIntel = true;
			break;
		case VIA:
			bridgeDev = new AGPVIA(provider);
			if((str = bridgeDev->Match(deviceID)) == NULL)
			{
				bridgeDev->release();
				bridgeDev = new AGPAMD64(provider);
				str = bridgeDev->Match(deviceID);
			}
			break;
		case SiS:
		case AMD64:
			bridgeDev = new AGPAMD64(provider);
			str = bridgeDev->Match(deviceID);
			break;

		default:
			str = NULL;
	}
	
	if(str != NULL)
	{
		IOLog("AGP: Found %s\n ", str);  
		bridgeDevice = bridgeDev->bridge;
	}
	else
	{
		bridgeDev->release();
	//	IOLog("AGP: not supported %04x\n", deviceID);
		return NULL;
	}
	bridgeDev->printBridgeConfig((IOPCIDevice *)provider);

	//provider->setProperty("AGP_Target", (void *) "", 0);
//	UInt32 devsel = 1;
	provider->setProperty("devsel-speed", &bridgeDev->devsel, sizeof(bridgeDev->devsel));

	*score 		-= 100;
	
	return (this);
}

//----------------------------------------------------------------------------

bool 
AGPGart::start( IOService * provider )
{ 
	bool ok;
	
    reserved = IONew(ExpansionData, 1);
    if (reserved == 0) return (false);
	
    bzero(reserved, sizeof(ExpansionData));
//	IOLog("AGPGart: start with %s\n", provider->getName());
/*	
	if (0 == (bridgeDev->bridge = OSDynamicCast(IOPCIDevice, provider)))
		return (0);
	
	bridgeDevice = bridgeDev->bridge;
*/	
    ok = super::start(provider);
	
//    if (ok && fBridgeInterruptSource) 		changePowerStateTo(2);

    return (ok);
	
/*IOLog("AGPGart: start\n");
	if(!(provider->getProperty("AGP_Target"))) return( super::start( provider)); 	
		IOLog("AGPGart: start with %s\n", provider->getName());
		
		if (0 == (bridgeDev->bridge = OSDynamicCast(IOPCIDevice, provider)))
			return (0);
	
		bridgeDevice = bridgeDev->bridge;
		
	//bridgeDev->printBridgeConfig((IOPCIDevice *)provider);
			
	return( super::start( provider)); 	*/
}

//----------------------------------------------------------------------------

bool 
AGPGart::configure( IOService * provider )
{
	UInt32 start, end;
	bool ok;
	IOPCIAddressSpace	space;
//	if(!provider->getProperty("AGP_Target")) return (super::configure(provider));
//	if(!(provider->getProperty("devsel-speed"))) return( super::configure( provider)); 	
		

	space = (bridgeDevice->space);
//	spaceFromProperties( provider, &space);
/*	IOPCIDevice *host = OSDynamicCast(IOPCIDevice, provider);
	IOPCIAddressSpace spaceH=host->space;
	UInt32 tmp=host->configRead32(spaceH,0);
	if((tmp == 0x59501002) || (tmp == 0x79301002))
	{
		IOLog("AGPATI trace PCI space AGP Bridge dev=1\n");
		for(int i=0; i<0x4f; i+=16)
		{
			for(int j=0; j<15; j+=4)
				IOLog("(%02lx)=%08lx   ",(UInt32)(i+j), (UInt32)host->configRead32(spaceH,i+j));  //for trace
			IOLog("\n");
		} 
		spaceH.s.deviceNum = 0;
		
		IOLog("AGPATI trace PCI space Host dev=0\n");
		for(int i=0; i<0xff; i+=16)
		{
			for(int j=0; j<15; j+=4)
				IOLog("(%02lx)=%08lx   ",(UInt32)(i+j), (UInt32)host->configRead32(spaceH,i+j));  //for trace
			IOLog("\n");
		} 
	}	
	*/
	space.s.deviceNum = 0;

	if( !(findPCICapability( space, kIOPCIAGPCapability, (UInt8 *)&bridgeDev->targetAGPRegisters )))
	{
		if(!(findPCICapability( space, kIOPCIAGP8Capability, (UInt8 *)&bridgeDev->targetAGPRegisters )))
		{
			IOLog("AGP: ERROR! Capability offset not found\n"); 
			return (false);
		} else
		IOLog("AGP: found AGP8 !!!\n");
	}
//		IOLog("AGPGart: configure %s\n", provider->getName());

//	UInt8 targetAGPRegisters = (UInt8)bridgeDev->targetAGPRegisters;
//	IOLog("AGP: capptr = %02x\n", (UInt8)bridgeDev->targetAGPRegisters);
    secondaryBus = bridgeDevice->configRead8( kPCI2PCISecondaryBus );
//	int priBus = bridgeDevice->configRead8( kPCI2PCIPrimaryBus );
    if (!secondaryBus)
    {
		IOLog("AGPBridge has secondaryBus 0\n");
		return (false);
    }
	subordinateBus = bridgeDevice->configRead8( kPCI2PCISubordinateBus );
//	IOLog("AGPBridge buses: pri %d secBus %d subBus %d BridgeDev %d\n",priBus, secondaryBus, subordinateBus, space.s.busNum);
	if (subordinateBus != secondaryBus)
	{
		subordinateBus = secondaryBus;
		bridgeDevice->configWrite8( kPCI2PCISubordinateBus, subordinateBus );
		IOLog("AGPGart subBus corrected to %d\n", subordinateBus);
	}
//Slice - all further from AGP@1
    end = bridgeDevice->configRead32( kPCI2PCIMemoryRange );
    if (end)
    {
        start = (end & 0xfff0) << 16;
        end |= 0x000fffff;
        ok = addBridgeMemoryRange( start, end - start + 1, false );  //Slice - were false
//		IOLog("AGPgart: BridgeMemoryRange [%08lx, %08lx]\n", start, end);
    }
	end = bridgeDevice->configRead32( kPCI2PCIPrefetchMemoryRange );
    if (end)
    {
        start = (end & 0xfff0) << 16;
		bridgeBase = start;
		end |= 0x000fffff;
		bridgeSize = end - start + 1;
        ok = addBridgePrefetchableMemoryRange( start, bridgeSize,  true );  //Slice - were false
//		IOLog("AGPgart: BridgePrefetchableMemoryRange [%08lx, %08lx]\n", start, end);
    }
//	ok = addBridgePrefetchableMemoryRange( 0x38000000, 0x40000000,  true );  //Slice - were false
//	IOLog("AGPgart: BridgePrefetchableMemoryRange [%08lx, %08lx]\n", 0x38000000, 0x40000000);
	
    end = bridgeDevice->configRead32( kPCI2PCIIORange );
    if (end)
    {
        start = (end & 0xf0) << 8;
        end = (end & 0xffff) | 0xfff;
        ok = addBridgeIORange( start, end - start + 1 );
//		IOLog("AGPgart: BridgeIORange [%08lx, %08lx]\n", start, end);
    }
	
	// enable card's PCI response
    bridgeDevice->setMemoryEnable(true);
    bridgeDevice->setIOEnable(true);
    bridgeDevice->setBusMasterEnable(true);
	
	
    saveBridgeState(); 
	
    if (bridgeDevice->savedConfig)
	{
        configShadow(bridgeDevice)->flags |= kIOPCIConfigShadowBridge; 
		configShadow(bridgeDevice)->bridge = (IOPCI2PCIBridge *)this;
	}
//	IOLog("AGP configure end\n");
//Slice - crash here
//	return (super::configure(provider));
	return true;
}

//--------------------------------------------------------------------
//Slice do not rewrite getBridgeSpace !!!

//----------------------------------------------------------------------------

IOPCIAddressSpace 
AGPGart::getBridgeSpace( void )
{
//	IOPCIAddressSpace	space;
//	IOLog("AGPGart: getBridgeSpace\n");
//	space = (bridgeDevice->space);

/*    space.bits = 0;
    space.s.busNum = secondaryBus;
	space.s.deviceNum = 0;
    return( space ); */
	return (bridgeDevice->space);
}

//----------------------------------------------------------------------------

IOPCIDevice * 
AGPGart::createNub( OSDictionary * from )
{
    IOPCIDevice *	nub;
	UInt8		masterReg;
	IOPCIAddressSpace displayDev;
    bool		isAGP;

    spaceFromProperties( from, &displayDev);
	isAGP = ( /* (displayDev.s.deviceNum != getBridgeSpace().s.deviceNum) && */
	     findPCICapability( displayDev, kIOPCIAGPCapability, &masterReg ));
if(isAGP) IOLog("AGP create nub for AGP Card=%08lx bus=%d cap=%d\n", configRead32(displayDev, 0), displayDev.s.busNum, masterReg);


	
	if(isAGP && ((pci_get_class(displayDev) == PCIC_DISPLAY) &&
		(pci_get_subclass(displayDev) == PCIS_DISPLAY_VGA)) )
	{
		//if (findPCICapability( displayDev, kIOPCIAGPCapability, &masterReg ))
		//{
	//	if(isAGP)
	//	{
			nub = new IOAGPDevice;
		
			if (nub)
				((IOAGPDevice *)nub)->masterAGPRegisters = masterReg;					
			// Those IOAGPFlags are documented in IOAGPDevice.h. However some bits are missing.
			// On PPC, a value of 65541 means 0x10005, we have the documentation for the low bits 
			// flags but what about	the 0x10000? After testing, the performance is the same with
			// and without, so we use a default value of '1'.
			
			// ON X1600, setting those flags to 1 makes the GUI boot, 
			// but once in GUI, nothing is cached in the GART. It still use the PCI GART in
			// the ATI driver. Setting it to 0, makes a bluescreen and some
			// people reported they made it to the GUI, but with really weird 
			// graphics, then the system hangs (panic).
			/* --------------------------------------------------------------------- 
			* #define kIOAGPBusFlagsKey	"IOAGPFlags"
			* 
			* // the AGP target must be idle before invalidating its gart tlb
			* kIOAGPGartIdleInvalidate	= 0x00000001,
			*
			* // the AGP target cannot handle operations that cross page boundaries
			* kIOAGPDisablePageSpans	= 0x00000002,
			*
			* // the AGP target cannot handle master -> target AGP writes
			* kIOAGPDisableAGPWrites	= 0x00000004,
			*
			* // the AGP target cannot handle target -> master PCI reads
			* kIOAGPDisablePCIReads		= 0x00000008,
			*
			* // the AGP target cannot handle master -> target PCI writes
			* kIOAGPDisablePCIWrites	= 0x00000010,
			*
			* // the AGP target cannot handle all unaligned transactions 
			* kIOAGPDisableUnaligned	= 0x00000020,
			*
			* kIOAGPDisableFeature6		= 0x00000040,
			* kIOAGPDisableFeature7		= 0x00000080,
			* kIOAGPDisableFeature8		= 0x00000100,
			* kIOAGPDisableFeature9		= 0x00000200
			* --------------------------------------------------------------------- */
	/*		UInt64	   flags;
			OSNumber * num;				
			
			flags = 0x10027;  //joblo - 65541=0x10005 Apple - 65543=0x10007
			num = OSNumber::withNumber(flags, 64);
				
			if (num)
			{
				from->setObject(kIOAGPBusFlagsKey, num);
				num->release();
			} */
			OSNumber * AGPMode = (OSNumber *)getProperty("AGP_Mode"); // kIOAGPBusFlagsKey
			if(AGPMode)
			{
				flags = AGPMode->unsigned32BitValue(); 
				//AGPMode->release();  //Slice - kernel panic
			}
			OSNumber * AGPBase = (OSNumber *)getProperty("HOST_Base");
			if(AGPBase)
			{
				customBase = AGPBase->unsigned32BitValue(); 
				//AGPBase->release();
			}
			OSNumber * FlushTime = (OSNumber *)getProperty("AGP_FlushTimeout"); 
			if(FlushTime)
			{
				bridgeDev->flusht = FlushTime->unsigned32BitValue(); 
				//FlushTime->release();
			}
			from->setObject( kIOAGPBusFlagsKey, getProperty( kIOAGPBusFlagsKey ) );
		//	return ( nub );
//		}
	}
	    else
        nub = super::createNub( from );

    return( nub );

	//Not an AGP device, create an IOPCIDevice. At this point, only ATI cards with 
	//dual-heads will reach this. 
	//return super::createNub( from );
}

//----------------------------------------------------------------------------
#define kIOFBBootGrayValue		0x00648cc3
#define INNUB(offset) OSReadLittleInt32((nubio_base), offset)
#define OUTNUB(offset,val) OSWriteLittleInt32((nubio_base), offset, val)


bool 
AGPGart::publishNub( IOPCIDevice * nub, UInt32 index )
{
	//UInt32	nubio_base_phys;
//	IOMemoryMap	*nubio;
//	volatile UInt8 *nubio_base;
    if (nub)
        nub->setProperty( "IOChildIndex" , index, 32 );

	// I subclass this method as this is the earliest point where we can
	// set various IORegistry entries, right before the IOFramebuffer class 
	// comes into play.

	if(super::publishNub(nub, index))
	{
		
		//nub->setIOEnable(true);
		//space = (ATIDevice->space);
/*		IOLog("AGP trace nub space\n");
		for(int i=0; i<0x6f; i+=16)
			{
				for(int j=0; j<15; j+=4)
					IOLog("(%02lx)=%08lx   ",(UInt32)(i+j), (UInt32)nub->configRead32(i+j));  //for trace
				IOLog("\n");
			}*/ 
		UInt32 ClassVGA = nub->configRead16(0x0a);
		if(!((ClassVGA & 0xffff) == 0x0300)) return true;
		UInt32 overlap = 1;
		UInt64 range = 0x00000000ffffffffULL;
	//	IOLog("AGPGart: publishNub %s\n", nub->getName());
		//Slice - these properties really not used
		nub->setProperty("AGP_Master", (void *) "", 0); 
		nub->setProperty("AGP_AllowOverlap", &overlap, sizeof(overlap)); //Slice - used only on PPC
		nub->setProperty("AGP_Address_Range", &range, sizeof(range)); //Slice - used only on PPC

	//	nub->setProperty( kIOAGPBusFlagsKey, &flags, sizeof(flags));
		
//		UInt32 VendVGA = nub->configRead32(0);
//		if(!((VendVGA  & 0xffff) == 0x1002))
		
		 return true;
#if 0		 
//		IOLog("AGPGart found VGA %04lx class %04lx\n", VendVGA, ClassVGA);
		nub->setMemoryEnable(true);

		/*for (UInt32 i = 0; (nubio = nub->mapDeviceMemoryWithIndex(i)); i++)
		{
			nubio_base_phys = nubio->getPhysicalAddress();
			// Make sure we  select MMIO registers
				if (((nubio->getLength()) == 0x00010000) && (nubio_base_phys != 0))
                break;
		}
		*/
		
		IOMemoryDescriptor *nubMemory = IOMemoryDescriptor::withPhysicalAddress(nub->configRead32(0x18), 0x1000, kIODirectionInOut);
		if(nubMemory)
		{
			nubio = nubMemory->map();
			if(!nubio)
			{
				IOLog("AGPGart: Radeon memory not mapped\n");
			} else 
			{
		//		mmio_base_phys = mmio->getPhysicalAddress();
				nubio_base = (volatile UInt8 *)nubio->getVirtualAddress();
				OUTNUB(0xf8, 0x04000000);
				OUTNUB(0x14c, 0x3fff3c00);
				IOLog("AGP: Radeon configured [3fff3c00]\n");
			}
		} else 
			IOLog("AGPGart: Radeon memory not created\n");
		return(true);
#endif		
	} else
		return (false);

}

#undef super
#define super IOPCIBridge
/**************************************************************************************/

IOReturn 
AGPGart::createAGPSpace( IOAGPDevice * master,	IOOptionBits options,
												IOPhysicalAddress * address,
												IOPhysicalLength * length )
{
	IOPCIAddressSpace target = getBridgeSpace();
	IOPhysicalLength agpLength	= *length;
	IOPhysicalAddress base;		//= *address; //Slice - this is not input parameter!
	IOReturn err = kIOReturnSuccess;
	// Get the AGP mode of operation.
	if((bridgeDev->GetAGPVersion(master)) > 1)
		IOLog("AGP: Found an AGP %d.%d compliant device.\n", bridgeDev->major_versionM, bridgeDev->minor_versionM );
	else
		return kIOReturnUnsupported;
//	if(bridgeSize < agpLength) 
		agpLength = 	bridgeSize;
	base = customBase;

	
	// Attach device. All the configuration is done through this method.
	// If we fail here, it's too late to recover the IOAccelerator but at
	// error out cleanly.
	if (bridgeDev->Configure(master, &base, &agpLength) != kIOReturnSuccess)
	{
		IOLog("AGP: ERROR! Attach device failed.");
		destroyAGPSpace( master );
		return kIOReturnVMError;
	}

	//Slice - apply bridge parameters to master
	master->setProperty("devsel-speed", &bridgeDev->devsel, sizeof(bridgeDev->devsel));
	if(bridgeDev->fastback)
	master->setProperty("fast-back-to-back", &bridgeDev->fastback, sizeof(bridgeDev->fastback));
	master->setProperty("max-latency", &bridgeDev->maxlat, sizeof(bridgeDev->maxlat));
	master->setProperty("min-grant", &bridgeDev->mingnt, sizeof(bridgeDev->mingnt));



		
	//IOLog("AGP: Aperture [%08lx:%08lx]\n",base, agpLength); 
	
	entries = agpLength >> AGP_PAGE_SHIFT;
	master->setProperty("AGP_Alignment", &agpLength, sizeof(agpLength));
	master->setProperty("AGP_Address_Block", &agpLength, sizeof(agpLength)); 
	//Slice - Create Gart
    do {
	
		// Create our GART table.
		ppnum_t physPage;
		UInt64 gartPhys64;
		int gw = bridgeDev->gartWidth >> 3;
		UInt32 gartLength = entries * gw + 0x2000;
		//IOLog("AGP: gartWidth=%d entries=%08lx entryWidth=%d\n",(int)bridgeDev->gartWidth, entries, gw);
		gartHandle = IOMapper::NewARTTable(gartLength, (void **) &gartArray, &physPage);
	 		//emersonfxbx method
		//gartArray = (volatile UInt32 *)IOMallocContiguous(entries * gw + 0x2000, 4096, &gartPhys);
		gartPhys64 = ptoa_64(physPage);
			
		if (!gartHandle) continue;
	//	IOSetProcessorCacheMode(kernel_task, (vm_address_t) gartArray, (entries * gw + 0x2000),
	//								/*kIOMapInhibitCachekIOMapCopybackCache*/kIOMapDefaultCache);

		bzero( (void *) gartArray, gartLength);
		flush_dcache( (vm_offset_t) gartArray, gartLength, false);

		gartPhys = gartPhys64 >> 12;
		//Really ranges is not used
        IORangeAllocator * platformRanges = getPlatform()->getPhysicalRangeAllocator();
		for( agpBaseIndex = 0xf; agpBaseIndex > 0; agpBaseIndex--)
		{
			customBase = agpBaseIndex * 0x10000000;
			if( platformRanges->allocateRange( customBase, agpLength ))
			{
				systemLength = agpLength;
				break;
			}
		}
		
		if( !systemLength)    continue;

		agpRange = IORangeAllocator::withRange( agpLength, 4096 );
		if ( !agpRange ) continue;

//		IODeviceMemory *	Vmem = bridgeDev->getVRAMMemory(master);
		
	//	*address = base; // - Vmem->getPhysicalAddress(); // 271a
		*length = systemLength;

		err = bridgeDev->InstallGATT(physPage);
		*address = systemLength;  //Slice - this is a great difference 270 from all other versions

		
   } while( false );
		
	// Write the AGPCMD register.
	
	if ( kIOReturnSuccess == err)
	{
		bridgeDev->Enable(master, flags);
	//	UInt32 gartWidth = 32; 
		//master->setProperty("AAPL,gart-width", &gartWidth, sizeof(gartWidth));
		OSNumber *num = OSNumber::withNumber(bridgeDev->gartWidth, 32);
		if(num)
		{
			master->setProperty("AAPL,gart-width", num);
			master->setProperty("NVDA,gart-width", num);
			num->release();
		} 
		master->masterState |= kIOAGPStateEnabled;
	} else
		destroyAGPSpace( master );
	
				
    return (err);
}

//----------------------------------------------------------------------------

IOReturn 
AGPGart::destroyAGPSpace( IOAGPDevice * master )
{
	bridgeDev->Disable( master );
	IOLog("AGPGart: destroyAGPSpace\n");	
	if( gartArray) {	//bridgeDev->
	
	IOMapper::FreeARTTable(gartHandle, entries * (bridgeDev->gartWidth >> 3) + 0x2000);
	//IOFreeContiguous( (void *)gartArray, entries * (bridgeDev->gartWidth >> 3) + 0x2000);	//gartArray, gartLength );

		gartHandle = 0;
		gartArray  = 0;
    }
	
	if ( bridgeDev->agpRange ) {
		bridgeDev->agpRange->release();
		bridgeDev->agpRange = 0;
	}

/*	if(agpMemory)
		agpMemory->release();
*/	
    if( bridgeDev->systemLength)
		bridgeDev->systemLength = 0;

    return( kIOReturnSuccess );
}

//----------------------------------------------------------------------------
static int IOLogCnt = 0;
static int cnt = 0;
//static UInt32 physOld = 0;

IOReturn 
AGPGart::commitAGPMemory( IOAGPDevice * master,	IOMemoryDescriptor * memory,
						  IOByteCount agpOffset,
						  IOOptionBits options )
{
	IOPCIAddressSpace 	target = getBridgeSpace();
    IOReturn		err = kIOReturnSuccess;
	IOByteCount	len, len2 = 0, gartOffset, gw;
	IOPhysicalAddress phys2=0; //	phys,
	IOByteCount         agpFlushStart;
//		, flushLength;
	addr64_t		phys64;
	UInt32		PTE, offset = 0;
	assert( agpOffset < systemLength );
	if(memory)
	{
		bridgeDev->Flush0TLB();
		//		pciwrite32(hostSpace, targetAGPRegisters + AGP_CTRL, 0);		
		
		phys2 = memory->getPhysicalAddress();
		/*		if(phys2 != physOld)
		{
			IOLogCnt = 0;
			physOld = phys2;
		} 
		*/
		len = memory->getLength();
		len2 = len;
		if(bridgeDev->gartWidth == 32)
		{
			gartOffset = agpOffset >> 10; gw = 4;
		} else {
			gartOffset = agpOffset >> 9; gw = 8;
		}
		
		agpFlushStart = gartOffset;
		while( (phys64 = memory->getPhysicalSegment64(offset, &len)))
		{
			offset += len;
			len = (len + PAGE_MASK) & ~PAGE_MASK;
	//		if(bridgeDev->isAMD){
				PTE = (phys64 & ~PAGE_MASK) | ((phys64 >> 28) & 0xff0);
	//		} else
	//			PTE = (phys64 & ~PAGE_MASK);
			//if(0) phys64 += agpOffset; //options
			while( len > 0) 
			{
				//phys |= bridgeDev->addrMask;
				//OSWriteLittleInt32(gartArray, gartOffset, 0);
				//(void)OSReadLittleInt32(gartArray, gartOffset);
				OSWriteLittleInt32(gartArray, gartOffset, PTE | bridgeDev->addrMask);
				//OUTGART(gartOffset, phys | bridgeDev->addrMask);
				//gartArray[gartOffset] = phys;
				gartOffset += gw;
				PTE += page_size;
				len -= page_size;
				
			}
		}	
		gartOffset -= gw;
		//phys2 = INGART(gartOffset);
		//Slice - it was up to 2.6.6, disable for Intel in 2.6.7
		/*		agpFlushStart &= ~31;
		flushLength = (gartOffset - agpFlushStart);
		flushLength = (flushLength + 31) & ~31;
		//    sync();
		//	isync();
		
		flush_dcache( ((vm_offset_t)gartArray)+agpFlushStart, flushLength, false);
		//		sync();
		//		isync();
		*/
		len = OSReadLittleInt32(gartArray, gartOffset);
		
		if(!bridgeDev->isAMD)
			// Deal with stupid Pentium 4 8-deeps store queue crap.
			for(offset = 0; offset < 64*16; offset += 64)
			{
				len = OSReadLittleInt32(gartArray, offset);
				OSWriteLittleInt32(gartArray, offset, len);
			}
				
				//		temp = pciread32(hostSpace, targetAGPRegisters + AGP_CTRL);
				//		pciwrite32(hostSpace, targetAGPRegisters + AGP_CTRL, temp | (1 << 7) | (1 << 8));		
				
				bridgeDev->Flush1TLB();
	}  else
	{
		IOLog("AGPGart: memory not prepared\n");
		err = kIOReturnError;
	}
	//agp_flush_cache();
	
	
	if (IOLogCnt < 2)
	{
		//	bridgeDev->CheckTLB();
		//	IOLog("AGP: commitAGPmemory options %02lx memoryPhys=%08lx\n", (UInt32)options, phys2);
		IOLog("AGP memory %08lx length %08lx offset %08lx cnt %08lx\n", //@gartPhys %08lx
			  (UInt32)phys2, (UInt32)len2, (UInt32)agpOffset, (UInt32)cnt);
		//	*(UInt32*)(gartPhys<<12+gartOffset), );
		cnt++;
		IOLogCnt++;
	}
	
	
	/*if(memory)
	{
		if(bridgeDev->BindMemory(memory,agpOffset) != 0)
			err = kIOReturnError;
	}	*/		
	return (err);
	
}

/* -------------------------------------------------------------------------- */
/* THE FIVE NEXT METHODS BELOW ARE OBSOLETE, THEY ARE NEVER CALLED			  */
/* -------------------------------------------------------------------------- */

IOReturn 
AGPGart::releaseAGPMemory( IOAGPDevice * master, IOMemoryDescriptor * memory,
												 IOByteCount agpOffset,
												 IOOptionBits options )
{
	IOLog("AGPGart: releaseAGPMemory\n");
		IOPCIAddressSpace 	target = getBridgeSpace();
    IOReturn		err = kIOReturnSuccess;
	IOByteCount	len,  gartOffset, gw; //len2,
//	IOPhysicalAddress	 phys2=0;
	
	if(memory)
	{
		bridgeDev->Flush0TLB();
		len = memory->getLength();
		if(bridgeDev->gartWidth == 32)
		{
			gartOffset = agpOffset >> 10; gw = 4;
		} else {
			gartOffset = agpOffset >> 9; gw = 8;
			}
		while( len > 0) 
		{
		gartArray[gartOffset++] = 0;
			len -= page_size;
		}
		//phys2 = OSReadLittleInt32(gartArray, gartOffset - gw);
		//phys2 = INGART(gartOffset - gw);
		bridgeDev->Flush1TLB();
	}  else
	{
		IOLog("AGPGart: memory not prepared\n");
		err = kIOReturnError;
	}

	return( kIOReturnSuccess );
}

//----------------------------------------------------------------------------

IORangeAllocator * 
AGPGart::getAGPRangeAllocator( IOAGPDevice * master )
{
//	IOLog("AGPGart: getAGPRange\n");
	return( bridgeDev->agpRange );
}

//----------------------------------------------------------------------------

IOOptionBits 
AGPGart::getAGPStatus( IOAGPDevice * master, IOOptionBits options )
{
	return( master->masterState );
}

//----------------------------------------------------------------------------

IOReturn 
AGPGart::resetAGPDevice( IOAGPDevice * master, IOOptionBits options )
{
    return( kIOReturnSuccess );
}

//----------------------------------------------------------------------------

IOReturn 
AGPGart::getAGPSpace( IOAGPDevice * master,	IOPhysicalAddress * address,
											IOPhysicalLength * length )
{
//	IOLog("AGPGart: getAGPSpace\n");
	if( bridgeDev->systemLength) {
		if (address)
			*address = bridgeDev->systemBase;
		if (length)
			*length = bridgeDev->systemLength;
		return( kIOReturnSuccess );
	} else
		return (kIOReturnNotReady);	
}

//----------------------------------------------------------------------------
//Slice - TODO look into AppleMacRiscPCI and take into accont
// ATI private registers.
/* POWER MANAGEMENT */

IOReturn 
AGPGart::saveDeviceState( IOPCIDevice * device,
                                           IOOptionBits options )
{
    IOReturn		ret;
    IOAGPDevice *	agpDev;
    UInt32		agpSave[3] = {0, 0, 0};
    IOPCIAddressSpace 	target = getBridgeSpace();
	target.s.deviceNum = 0;

    if( (agpDev = OSDynamicCast( IOAGPDevice, device)))
	 {
		if( agpDev->masterState & kIOAGPStateEnabled)
            bridgeDev->Disable( agpDev );
		agpSave[0] = configRead32( target, kiAPBASE );
        agpSave[1] = configRead32( target, bridgeDev->targetAGPRegisters + AGP_ATTBASE );
        agpSave[2] = configRead32( target, bridgeDev->targetAGPRegisters + kIOPCIConfigAGPCommandOffset );
    }

    ret = super::saveDeviceState( device, options);

   if( agpDev && (ret == kIOReturnSuccess)) {
        agpDev->savedConfig[ 16 ] = agpSave[0];
        agpDev->savedConfig[ 17 ] = agpSave[1];
        agpDev->savedConfig[ 18 ] = agpSave[2];
    }

    return( ret );
}

//----------------------------------------------------------------------------

IOReturn 
AGPGart::restoreDeviceState( IOPCIDevice * device,
                                              IOOptionBits options )
{
    IOReturn		ret;
    IOAGPDevice *	agpDev;
    UInt32		agpSave[3] = {0, 0, 0};
    IOPCIAddressSpace 	target = getBridgeSpace();
	target.s.deviceNum = 0;

    agpDev = OSDynamicCast( IOAGPDevice, device);
    if( agpDev && device->savedConfig) {
        agpSave[0] = agpDev->savedConfig[ 16 ];
        agpSave[1] = agpDev->savedConfig[ 17 ];
        agpSave[2] = agpDev->savedConfig[ 18 ];
    }

    ret = super::restoreDeviceState( device, options);

    if( agpDev && (kIOReturnSuccess == ret)) {
        configWrite32( target, kiAPBASE, agpSave[0] );
        configWrite32( target, bridgeDev->targetAGPRegisters + AGP_ATTBASE , agpSave[1] );

		/*if(bridgeDev->readyForPM)
			bridgeDev->Enable(agpDev, flags);
		*/
        if( kIOAGPEnable & agpSave[2])
            agpDev->masterState |= kIOAGPStateEnablePending;
        else
            agpDev->masterState &= ~kIOAGPStateEnablePending;
    }
    
    return( ret );
}

//----------------------------------------------------------------------------

int 
AGPGart::GetVendor(IOService *provider)
{
	OSData *		data;
	UInt16 vendID;
	UInt16 vendor;
	
	/* Vendor ID's of current prober */
	data = OSDynamicCast(OSData, provider->getProperty(kVendor_IDProperty));
	vendID = *((UInt16 *) data->getBytesNoCopy());
	//IOLog("AGP: Vendor = %04x\n", vendID);
	switch (vendID) 
	{
		case 0x1106:
				vendor = VIA;				
			break;
		case 0x1039: /* SIS */
				vendor = SiS;
			break;
		case 0x1022: /*	AMD */
		case 0x10b9: /* ULI */
		case 0x10de: /* NVIDIA */
				vendor = AMD64;
			break;
		case 0x1002:	/* ATI */
				vendor = ATI;
			break;
		case 0x8086:	/* INTEL */
				vendor = INTEL;
			break;
		default:
			return -1;
	}
	return vendor;
}
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

void AGPGart::saveBridgeState( void )
{
    long cnt;

    for (cnt = 0; cnt < kIOPCIBridgeRegs; cnt++)
    {
        bridgeState[cnt] = bridgeDevice->configRead32(cnt * 4);
    }
//	IOLog("AGP: saveBridgeState OK\n");
}

void AGPGart::restoreBridgeState( void )
{
  long cnt;
  
    // start at config space location 8 -- bytes 0-3 are
    // defined by the PCI Spec. as ReadOnly, and we don't
    // want to write anything to the Command or Status
    // registers until the rest of config space is set up.
//IOLog("AGP: restoreBridgeState\n");
    for (cnt = (kIOPCIConfigCommand >> 2) + 1; cnt < kIOPCIBridgeRegs; cnt++)
    {
        bridgeDevice->configWrite32(cnt * 4, bridgeState[cnt]);
    }

    // once the rest of the config space is restored,
    // turn on all the enables (,etc.) in the Command register.
    // NOTE - we also reset any status bits in the Status register
    // that may have been previously indicated by writing a '1'
    // to the bits indicating whatever they were indicating.

    bridgeDevice->configWrite32(kIOPCIConfigCommand,
				bridgeState[kIOPCIConfigCommand >> 2]);
//	IOLog("AGP: restoreBridgeState OK\n");
}


UInt8 AGPGart::firstBusNum( void )
{
    return bridgeDevice->configRead8( kPCI2PCISecondaryBus );
}

UInt8 AGPGart::lastBusNum( void )
{
    return bridgeDevice->configRead8( kPCI2PCISubordinateBus );
}


UInt32 AGPGart::configRead32( IOPCIAddressSpace space,
                                      UInt8 offset )
{
    return (bridgeDevice->configRead32(space, offset));
}

void AGPGart::configWrite32( IOPCIAddressSpace space,
                                     UInt8 offset, UInt32 data )
{
    bridgeDevice->configWrite32( space, offset, data );
}

UInt16 AGPGart::configRead16( IOPCIAddressSpace space,
                                      UInt8 offset )
{
    return (bridgeDevice->configRead16(space, offset));
}

void AGPGart::configWrite16( IOPCIAddressSpace space,
                                     UInt8 offset, UInt16 data )
{
    bridgeDevice->configWrite16( space, offset, data );
}

UInt8 AGPGart::configRead8( IOPCIAddressSpace space,
                                    UInt8 offset )
{
    return (bridgeDevice->configRead8(space, offset));
}

void AGPGart::configWrite8( IOPCIAddressSpace space,
                                    UInt8 offset, UInt8 data )
{
    bridgeDevice->configWrite8( space, offset, data );
}

IODeviceMemory * AGPGart::ioDeviceMemory( void )
{
    return (bridgeDevice->ioDeviceMemory());
}