/*
 * Copyright (c) 1998-2000 Apple Computer, Inc. All rights reserved.
 *
 * @APPLE_LICENSE_HEADER_START@
 * 
 * The contents of this file constitute Original Code as defined in and
 * are subject to the Apple Public Source License Version 1.1 (the
 * "License").  You may not use this file except in compliance with the
 * License.  Please obtain a copy of the License at
 * http://www.apple.com/publicsource and read it before using this file.
 * 
 * This Original Code and all software distributed under the License are
 * distributed on an "AS IS" basis, WITHOUT WARRANTY OF ANY KIND, EITHER
 * EXPRESS OR IMPLIED, AND APPLE HEREBY DISCLAIMS ALL SUCH WARRANTIES,
 * INCLUDING WITHOUT LIMITATION, ANY WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE OR NON-INFRINGEMENT.  Please see the
 * License for the specific language governing rights and limitations
 * under the License.
 * 
 * @APPLE_LICENSE_HEADER_END@
 */

#ifndef _IOKIT_AGPGart_H
#define _IOKIT_AGPGart_H


//#include <IOKit/IOLib.h>
#include <IOKit/pci/IOPCIBridge.h>
#include "pci/IOPCIPrivate.h"
#include "IOAGPBridge.h"

//#define kIOPCIBridgeRegs (0x20)

class IORangeAllocator;

class AGPGart : public IOPCIBridge
{
    OSDeclareDefaultStructors(AGPGart)

#define pci_get_cachelnsz(dev)   configRead8(dev, PCIR_CACHELNSZ)
#define pci_get_class(dev)       configRead8(dev, PCIR_CLASS)
#define pci_get_cmdreg(dev)      configRead16(dev, PCIR_COMMAND)
#define pci_get_device(dev)      configRead16(dev, PCIR_DEVICE)
#define pci_get_hdrtype(dev)     configRead8(dev, PCIR_HDRTYPE)
#define pci_get_intpin(dev)      configRead8(dev, PCIR_INTPIN)
#define pci_get_irq(dev)         configRead8(dev, PCIR_INTLINE)
#define pci_get_lattimer(dev)    configRead8(dev, PCIR_LATTIMER)
#define pci_get_maxlat(dev)      configRead8(dev, PCIR_MAXLAT)
#define pci_get_mingnt(dev)      configRead8(dev, PCIR_MINGNT)
#define pci_get_progif(dev)      configRead8(dev, PCIR_PROGIF)
#define pci_get_revid(dev)       configRead8(dev, PCIR_REVID)
#define pci_get_statreg(dev)     configRead16(dev, PCIR_STATUS)
#define pci_get_subclass(dev)    configRead8(dev, PCIR_SUBCLASS)
#define pci_get_subdevice(dev)   configRead16(dev, PCIR_SUBDEV_0)
#define pci_get_subvendor(dev)   configRead16(dev, PCIR_SUBVEND_0)
#define pci_get_vendor(dev)      configRead16(dev, PCIR_VENDOR)

public:
	volatile UInt32*			gartArray;
	UInt32						gartPhys;
	OSData *					gartHandle;

    IORangeAllocator *	agpRange;
    UInt32		agpBaseIndex;
	IOPhysicalAddress	customBase;
    IOPhysicalLength	systemLength;
	IOAGPBridge *bridgeDev;
	UInt8	vendor;
	
			UInt32			bridgeBase;
			UInt32			bridgeSize;
//			UInt32			flusht;
	UInt32		flags;
	UInt32			entries;
/*	IOMemoryDescriptor *agpMemory;
			 IOMemoryMap	*agpio;
		volatile UInt8 *agpio_base;
		unsigned long   agpio_base_phys; */

	IOPCIDevice		*bridgeDevice;
	UInt32        bridgeState[kIOPCIBridgeRegs];
	
protected:
    IOSimpleLock *		lock;
    IODeviceMemory *		ioMemory;
    IOMemoryMap *		configAddrMap;
    IOMemoryMap *		configDataMap;

    volatile UInt32	*	configAddr;
    volatile UInt8	*	configData;

    UInt16			coarseAddressMask;
    UInt16			fineAddressMask;

    UInt8			secondaryBus;
    UInt8			subordinateBus;
//    UInt8			configDataOffsetMask;
//    UInt32			uniNVersion;

//    inline bool setConfigSpace( IOPCIAddressSpace space, UInt8 offset );
    virtual UInt8 firstBusNum( void );
    virtual UInt8 lastBusNum( void );
	

public:
	virtual IOService * probe( IOService * provider, SInt32 * score );
	
    virtual bool start(	IOService * provider );
	
    virtual bool configure( IOService * provider );
	
    virtual IOPCIAddressSpace getBridgeSpace( void );
	
    virtual IOPCIDevice * createNub( OSDictionary * from );
	
	virtual bool publishNub( IOPCIDevice * nub, UInt32 index );
	
    virtual IODeviceMemory * ioDeviceMemory( void );

    virtual void saveBridgeState( void );
    virtual void restoreBridgeState( void );

    virtual UInt32 configRead32( IOPCIAddressSpace space, UInt8 offset );
    virtual void configWrite32( IOPCIAddressSpace space,
					UInt8 offset, UInt32 data );
    virtual UInt16 configRead16( IOPCIAddressSpace space, UInt8 offset );
    virtual void configWrite16( IOPCIAddressSpace space,
					UInt8 offset, UInt16 data );
    virtual UInt8 configRead8( IOPCIAddressSpace space, UInt8 offset );
    virtual void configWrite8( IOPCIAddressSpace space,
					UInt8 offset, UInt8 data );

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * */

    virtual IOReturn createAGPSpace( IOAGPDevice * master,
				     IOOptionBits options,
				     IOPhysicalAddress * address, 
				     IOPhysicalLength * length );

    virtual IOReturn destroyAGPSpace( IOAGPDevice * master );

    virtual IORangeAllocator * getAGPRangeAllocator( IOAGPDevice * master );

    virtual IOOptionBits getAGPStatus( IOAGPDevice * master,
				       IOOptionBits options );
    virtual IOReturn resetAGPDevice( IOAGPDevice * master,
                                     IOOptionBits options );

    virtual IOReturn getAGPSpace( IOAGPDevice * master,
                                  IOPhysicalAddress * address, 
				  IOPhysicalLength * length );

    virtual IOReturn commitAGPMemory( IOAGPDevice * master, 
						IOMemoryDescriptor * memory,
						IOByteCount agpOffset,
						IOOptionBits options );

    virtual IOReturn releaseAGPMemory(  IOAGPDevice * master, 
				      	IOMemoryDescriptor * memory, 
						IOByteCount agpOffset,
						IOOptionBits options );
					
	virtual IOReturn saveDeviceState( IOPCIDevice * device,
                                      IOOptionBits options );
    virtual IOReturn restoreDeviceState( IOPCIDevice * device,
                                         IOOptionBits options);
										 
	private:
		int GetVendor(IOService *provider);

};

#endif /* ! _IOKIT_AGPGart_H */

