
#include "IOAGPBridge.h"
#include "agpreg.h"
#include "agpio.h"

#define AGP_MAJOR_VERSION_SHIFT (20)
#define AGP_MINOR_VERSION_SHIFT (16)
#define kIOAGPCommandValueKey	"IOAGPCommandValue"

OSDefineMetaClassAndStructors(IOAGPBridge, OSObject)

//----------------------------------------------------------------------------
const char* IOAGPBridge::Match(UInt16 dev)
{
	return ( NULL );
}

void IOAGPBridge::printBridgeConfig(IOPCIDevice * AGPBridge)
{	
/*
	UInt8 first, last, cacheline, mingnt, lattimer, maxlat, irq;
	UInt16 stat, secstat, cmd;
	UInt16 base, limit, pbase, plimit;
	UInt8 iobase, iolimit, devsel, fastback;
*/	
	// PCI STATUS AND COMMAND REGISTERS
	stat		=	AGPBridge->configRead16(PCIR_STATUS);
	secstat		=	AGPBridge->configRead16(0x1e);
	cmd			=	AGPBridge->configRead16(PCIR_COMMAND);
	devsel		=	(stat & 0x600) >> 9;
	fastback	=	(stat & 0x80);
	//SECONDARY AND SUBORDINATE BUS NUMBERS
	first		=	AGPBridge->configRead8( kPCI2PCISecondaryBus );
	last		=	AGPBridge->configRead8( kPCI2PCISubordinateBus );
	// CACHE LINE SIZE
	cacheline	=	AGPBridge->configRead8( PCIR_CACHELNSZ );
	// MIN GRANT
	mingnt		=	AGPBridge->configRead8( PCIR_MINGNT );
	//LATENCY
	lattimer	=	AGPBridge->configRead8( PCIR_LATTIMER );
	maxlat		=	AGPBridge->configRead8( PCIR_MAXLAT );
	//IRQ
	irq			=	AGPBridge->configRead8( PCIR_INTLINE );
	//IO RANGES
	iobase		=	AGPBridge->configRead8(0x1c);
	iolimit		=	AGPBridge->configRead8(0x1d);
	//MEMORY RANGES
	base		=	AGPBridge->configRead16(0x20);
	limit		=	AGPBridge->configRead16(0x22);
	//PREFETCH RANGES
	pbase		=	AGPBridge->configRead16(0x24);
	plimit		=	AGPBridge->configRead16(0x26);
/*
	IOLog("--- PCI-TO-AGP BRIDGE CONFIGURATION ---\n");
	IOLog("PCI : STATUS = 0x%04x, SECSTATUS = 0x%04x, COMMAND = 0x%04x\n", stat, secstat, cmd);
	IOLog("BUSSES : FIRST = %d, LAST = %d\n", first, last);
	IOLog("LATENCY : TIMER = 0x%02x, MAX = 0x%02x\n", lattimer, maxlat);
	IOLog("IRQ : LINE = %d\n",irq);
	IOLog("IO RANGES : BASE = 0x%02x, LIMIT = 0x%02x\n", iobase, iolimit);
	IOLog("MEMORY RANGES : BASE = 0x%04x, LIMIT = 0x%04x\n", base, limit);
	IOLog("PREFETCH MEMORY RANGES : BASE = 0x%04x, LIMIT = 0x%04x\n",pbase, plimit);
	IOLog("Cache line size = 0x%02x, min grant = 0x%02x, devsel = %d, fast-back-to-back = %d\n", cacheline, mingnt, devsel, fastback);
	IOLog("---				DONE			  ---\n");
*/	
	
	
}

IODeviceMemory *IOAGPBridge::getVRAMMemory(IOAGPDevice * master)
{
	IODeviceMemory *	mem = 0;
	UInt32 maxLen = 0, idx = 0;
	
	for (UInt32 i = 0; (mem = master->getDeviceMemoryWithIndex(i)); i++)
	{
		// Make sure we don't select MMIO registers
//		if ((mem->getLength()) > 0x00200000)
//                break;
		if (mem->getLength() > maxLen)
		{
			idx=i;
			maxLen = mem->getLength();
		}
	}
	mem = master->getDeviceMemoryWithIndex(idx);
	IOLog("AGP: VRAM=[%08lx, %08lx]\n", mem->getPhysicalAddress(), mem->getLength());
	return (mem);

}

//----------------------------------------------------------------------------

IOReturn IOAGPBridge::Configure(IOAGPDevice	*master, 
						IOPhysicalAddress * address, IOPhysicalLength * length)
{
	return ( kIOReturnUnsupported );
}

//----------------------------------------------------------------------------


IOReturn IOAGPBridge::InstallGATT(IOPhysicalAddress gattPhys)
{
	return ( kIOReturnUnsupported );
}

//----------------------------------------------------------------------------

void IOAGPBridge::Flush0TLB()
{
	return;
}
//----------------------------------------------------------------------------

void IOAGPBridge::Flush1TLB()
{
	return;
}
//----------------------------------------------------------------------------

void IOAGPBridge::CheckTLB()
{
	return;
}

//----------------------------------------------------------------------------
/*
IOReturn IOAGPBridge::BindMemory(IOMemoryDescriptor * mem, IOByteCount agpOffset)
{
	return ( kIOReturnUnsupported );
}
*/
//----------------------------------------------------------------------------

void IOAGPBridge::Disable( IOAGPDevice * master )
{
	master->configWrite32(master->masterAGPRegisters + kIOPCIConfigAGPCommandOffset, 0 );
    pciwrite32( hostSpace, targetAGPRegisters + kIOPCIConfigAGPCommandOffset, 0 );

	master->masterState &= ~kIOAGPStateEnabled;
	
	return;
}

//----------------------------------------------------------------------------

int IOAGPBridge::GetAGPVersion(IOAGPDevice * master)
{
	UInt32 ncapidm;
	UInt8 masterAGPRegisters = master->masterAGPRegisters;
	
	ncapidm = master->configRead32(masterAGPRegisters);
	major_versionM = (ncapidm >> AGP_MAJOR_VERSION_SHIFT) & 0xf;
	minor_versionM = (ncapidm >> AGP_MINOR_VERSION_SHIFT) & 0xf;
	tstatus = pciread32(hostSpace, targetAGPRegisters + AGP_STATUS);
	mstatus = master->configRead32(masterAGPRegisters + AGP_STATUS);
	
	ITA_COH = (tstatus & (1<<8)) >> 7;
	if(tstatus & (1 << 7))
		gartWidth = 64;
	else
		gartWidth = 32;
	IOLog("AGP: Coherence support: %s\n", ITA_COH? "yes": "no");
	IOLog("AGP: GART is %d bit capable\n", (int)gartWidth);
	return major_versionM;
}

//----------------------------------------------------------------------------

IOReturn IOAGPBridge::Enable(IOAGPDevice * master, UInt32 mode)
{	
	//UInt32 mode	= flags; //0xFFFFFFFF;
/*	IOPCIAddressSpace masterSpace = master->space;
	UInt8 masterAGPRegisters	= master->masterAGPRegisters;


	tstatus = pciread32(hostSpace, targetAGPRegisters + AGP_STATUS);
	mstatus = master->configRead32(masterAGPRegisters + AGP_STATUS);
*/
	/*
	 * Check display and bridge for AGP v3 support.  AGP v3 allows
	 * more variety in topology than v2, e.g. multiple AGP devices
	 * attached to one bridge, or multiple AGP bridges in one
	 * system.  This doesn't attempt to address those situations,
	 * but should work fine for a classic single AGP slot system
	 * with AGP v3.
	 */
	if (major_versionM == 3)
		return (EnableMode3(master, mode));
	else
		return (EnableMode2(master, mode));	    
}

//----------------------------------------------------------------------------

IOReturn IOAGPBridge::EnableMode3(IOAGPDevice * agpMaster, UInt32 mode)
{
	UInt32 command;
	UInt32 rq, sba, fw, rate, arqsz, cal;
	UInt8 masterAGPRegisters = agpMaster->masterAGPRegisters;
	IOPCIAddressSpace masterSpace = agpMaster->space;
/*	UInt32 commandKey;

	commandKey = kIOAGPSideBandAddresssing | kIOAGP3Mode | 
			kIOAGP4xDataRate | kIOAGP2xDataRate | kIOAGP1xDataRate;
*/

	/* Set RQ to the min of mode, tstatus and mstatus */
	rq = AGP_MODE_GET_RQ(mode);
	if (AGP_MODE_GET_RQ(tstatus) < rq)
		rq = AGP_MODE_GET_RQ(tstatus);
	if (AGP_MODE_GET_RQ(mstatus) < rq)  //Slice: always 1
		rq = AGP_MODE_GET_RQ(mstatus);

	/*
	 * ARQSZ - Set the value to the maximum one.
	 * Don't allow the mode register to override values.
	 */
	arqsz = AGP_MODE_GET_ARQSZ(mode);
	if (AGP_MODE_GET_ARQSZ(tstatus) > rq)
		rq = AGP_MODE_GET_ARQSZ(tstatus);
	if (AGP_MODE_GET_ARQSZ(mstatus) > rq)	//Slice: always 0
		rq = AGP_MODE_GET_ARQSZ(mstatus);

	/* Calibration cycle - don't allow override by mode register */
	cal = AGP_MODE_GET_CAL(tstatus);
	if (AGP_MODE_GET_CAL(mstatus) < cal)
		cal = AGP_MODE_GET_CAL(mstatus);

	/* SBA must be supported for AGP v3. */
	sba = 1;

	/* Set FW if all three support it. */
	fw = (AGP_MODE_GET_FW(tstatus)
	       & AGP_MODE_GET_FW(mstatus)
	       & AGP_MODE_GET_FW(mode));
//	if(fw != 0) commandKey |= kIOAGPFastWrite;
	
	/* Figure out the max rate */
	rate = (AGP_MODE_GET_RATE(tstatus)
		& AGP_MODE_GET_RATE(mstatus)
		& AGP_MODE_GET_RATE(mode));
	if (rate & AGP_MODE_V3_RATE_8x)
		rate = AGP_MODE_V3_RATE_8x;
	else
		rate = AGP_MODE_V3_RATE_4x;
		IOLog("AGP: Setting %02lx data rate\n", rate * 4);

	/* Construct the new mode word and tell the hardware */
	command = AGP_MODE_SET_RQ(0, rq);
	command = AGP_MODE_SET_ARQSZ(command, arqsz);
	command = AGP_MODE_SET_CAL(command, cal);
	command = AGP_MODE_SET_SBA(command, 1);  //Slice AGP3 always SBA=1
	command = AGP_MODE_SET_FW(command, fw); //-- this is netkas recommendation fw=0
	command = AGP_MODE_SET_RATE(command, rate);
	command = AGP_MODE_SET_AGP(command, 1);
	if(gartWidth == 64) command |= (1<<7);
	
/*	commandKey &= tstatus;
	commandKey &= mstatus;
	
	commandKey &= 0xFFFFFFFF; */
	UInt32 tcom, mcom;
	//tcom = pciread32( hostSpace, targetAGPRegisters + kIOPCIConfigAGPCommandOffset);
	do {
		pciwrite32(hostSpace, targetAGPRegisters + kIOPCIConfigAGPCommandOffset, command);
		tcom = pciread32( hostSpace, targetAGPRegisters + kIOPCIConfigAGPCommandOffset);
	}  while( (command & kIOAGPEnable) != 	(kIOAGPEnable & tcom));
	
	mcom = agpMaster->configRead32( masterAGPRegisters + kIOPCIConfigAGPCommandOffset);
	do {
		agpMaster->configWrite32( masterAGPRegisters + kIOPCIConfigAGPCommandOffset, command );
		mcom = agpMaster->configRead32( masterAGPRegisters + kIOPCIConfigAGPCommandOffset);
	} while( (command & kIOAGPEnable) != (kIOAGPEnable & mcom));
	/*
	masterSpace.s.functionNum = 1;
	UInt8 cap2 = agpMaster->configRead8(masterSpace, PCIR_CAP_PTR);
	if(cap2)
	{
		UInt8 master2AGP = agpMaster->configRead8(masterSpace, cap2);
		do
		{
			agpMaster->configWrite32( masterSpace, master2AGP + kIOPCIConfigAGPCommandOffset, command );
		} while( (command & kIOAGPEnable) != 
			(kIOAGPEnable & agpMaster->configRead32(masterSpace, master2AGP + kIOPCIConfigAGPCommandOffset)));
	} 
	masterSpace.s.functionNum = 0;
	*/
	IOLog("AGP: command written target=%08lx master=%08lx\n", tcom, mcom);
	agpMaster->setProperty(kIOAGPCommandValueKey, &command, sizeof(command));
	agpMaster->setProperty("AAPL,agp3-mode", kOSBooleanTrue);
//	agpMaster->setProperty( kIOAGPBusFlagsKey, &AGPFlags, sizeof(AGPFlags) );


	return( kIOReturnSuccess );
	
}

//----------------------------------------------------------------------------

IOReturn IOAGPBridge::EnableMode2(IOAGPDevice * agpMaster, UInt32 mode)
{
	UInt32 command;
	UInt32 rq, sba, fw, rate;
	UInt8 masterAGPRegisters = agpMaster->masterAGPRegisters;
/*	UInt32 commandKey;

	commandKey = kIOAGPSideBandAddresssing | kIOAGP3Mode | 
			kIOAGP4xDataRate | kIOAGP2xDataRate | kIOAGP1xDataRate;
*/
	/* Set RQ to the min of mode, tstatus and mstatus */
	rq = AGP_MODE_GET_RQ(mode);
	if (AGP_MODE_GET_RQ(tstatus) < rq)
		rq = AGP_MODE_GET_RQ(tstatus);
	if (AGP_MODE_GET_RQ(mstatus) < rq)
		rq = AGP_MODE_GET_RQ(mstatus);

	/* Set SBA if all three can deal with SBA */
	sba = (AGP_MODE_GET_SBA(tstatus)
	       & AGP_MODE_GET_SBA(mstatus)
	       & AGP_MODE_GET_SBA(mode));

	/* Similar for FW */
	fw = (AGP_MODE_GET_FW(tstatus)
	       & AGP_MODE_GET_FW(mstatus)
	       & AGP_MODE_GET_FW(mode));
//	if(fw != 0)		commandKey |= kIOAGPFastWrite;

	/* Figure out the max rate */
	rate = (AGP_MODE_GET_RATE(tstatus)
		& AGP_MODE_GET_RATE(mstatus)
		& AGP_MODE_GET_RATE(mode));
	if (rate & AGP_MODE_V2_RATE_4x)
		rate = AGP_MODE_V2_RATE_4x;
	else if (rate & AGP_MODE_V2_RATE_2x)
		rate = AGP_MODE_V2_RATE_2x;
	else
		rate = AGP_MODE_V2_RATE_1x;

	IOLog("AGP: Device is in legacy mode, setting %02lx data rate\n", rate);

	/* Construct the new mode word and tell the hardware */
	command = AGP_MODE_SET_RQ(0, rq);
	command = AGP_MODE_SET_SBA(command, sba);
	command = AGP_MODE_SET_FW(command, fw);
	command = AGP_MODE_SET_RATE(command, rate);
	command = AGP_MODE_SET_AGP(command, 1);
	if(gartWidth == 64) command |= (1<<7);
	
	//commandKey &= tstatus;
	//commandKey &= mstatus;
	
	//commandKey &= 0xFFFFFFFF;
	UInt32 tcom, mcom;
	//tcom = pciread32( hostSpace, targetAGPRegisters + kIOPCIConfigAGPCommandOffset);
	do {
		pciwrite32(hostSpace, targetAGPRegisters + kIOPCIConfigAGPCommandOffset, command);
		tcom = pciread32( hostSpace, targetAGPRegisters + kIOPCIConfigAGPCommandOffset);
	}  while( (command & kIOAGPEnable) != 	(kIOAGPEnable & tcom));
	
	//mcom = agpMaster->configRead32( masterAGPRegisters + kIOPCIConfigAGPCommandOffset);
	do {
		agpMaster->configWrite32( masterAGPRegisters + kIOPCIConfigAGPCommandOffset,  command );
		mcom = agpMaster->configRead32( masterAGPRegisters + kIOPCIConfigAGPCommandOffset);
	} while( (command & kIOAGPEnable) != (kIOAGPEnable & mcom));
	
	agpMaster->setProperty(kIOAGPCommandValueKey, &command, sizeof(command));
	agpMaster->setProperty("AAPL,agp3-mode", kOSBooleanFalse);
	
	return( kIOReturnSuccess );
}

//----------------------------------------------------------------------------
