 /*-
 * Copyright (c) 2004, 2005 Jung-uk Kim <jkim@FreeBSD.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <IOKit/IOLib.h>
#include <IOKit/platform/ApplePlatformExpert.h>
#include <IOKit/pci/IOPCIBridge.h>
#include <IOKit/IOMapper.h>
#include <IOKit/IORangeAllocator.h>

#include "IOAGPBridge.h"

#define REG_GARTCTRL    0
#define REG_APSIZE      1
#define REG_ATTBASE     2

#define VIA_APSIZE      0x84
#define A_SIZE_8(x)     ((struct aper_size_info_8 *) x)

#define GPTE_VALID      1
#define GPTE_COHERENT   2


/********************************************************************/
/*						VIA DRIVER FUNCTIONS						*/
/*							by joblo10								*/
/*						 November 2nd, 2006,						*/
/*					  last update on 2006/11/24	
						Slice correct design for 2.6				*/	
/********************************************************************/

#define super IOAGPBridge
	
UInt8 via_v2_regs[] = { AGP_VIA_GARTCTRL, AGP_VIA_APSIZE, AGP_VIA_ATTBASE };
UInt8 via_v3_regs[] = { AGP3_VIA_GARTCTRL, AGP3_VIA_APSIZE, AGP3_VIA_ATTBASE };

//----------------------------------------------------------------------------

AGPVIA::AGPVIA(IOService *provider)
{
	//Accessor for configRead and configWrite 
	bridge = (IOPCIDevice *)provider;
	
	//PCI Space of the host bridge
	hostSpace = bridge->space;
	AGPBSpace = hostSpace;
	hostSpace.s.deviceNum = 0;	
}
/*
//----------------------------------------------------------------------------

IOReturn AGPVIA::BindMemory(IOMemoryDescriptor * mem, IOByteCount agpOffset)
{
	
   // UInt32		offset = 0;
	//UInt32	mem_phys;
	IOByteCount	len; //, len2;
	IOPhysicalAddress	phys, phys2;
		IOByteCount         agpFlushStart, flushLength;
// 	UInt32 temp;
	len = mem->getLength();
	phys = mem->getPhysicalAddress();

	UInt32 gartOffset = agpOffset >> 10;
	agpFlushStart = gartOffset;
	//agpOffset >> 10;	//	/= (page_size / 4);
		
//	while( (phys = mem->getPhysicalSegment(offset, &len)))
//	{
//		phys += agpOffset;
	//	offset += len;
	//	len = (len + PAGE_MASK) & ~PAGE_MASK;
		while( len > 0) 
		{
			OSWriteLittleInt32(gartArray, gartOffset, phys | addrMask);
			gartOffset += 4;
			phys += page_size;
			len -= page_size;
		}
//	}
	flushLength = (gartOffset - agpFlushStart);
	flush_dcache( ((vm_offset_t)gartArray)+agpFlushStart, flushLength, false);
	phys2 = OSReadLittleInt32(gartArray, gartOffset-4);
//	mem->performOperation(kIOMemoryIncoherentIOFlush, 0, len2); //mem->getLength() );
	agp_flush_cache();
	FlushTLB();

	return( 0 );
}
*/
//----------------------------------------------------------------------------

UInt32 AGPVIA::agp_via_get_aperture()
{
	struct aper_size_info_8 via_generic_sizes[7] =
	{
		{256, 65536, 6, 0},
		{128, 32768, 5, 128},
		{64, 16384, 4, 192},
		{32, 8192, 3, 224},
		{16, 4096, 2, 240},
		{8, 2048, 1, 248},
		{4, 1024, 0, 252}
	};
	struct aper_size_info_16 agp3_generic_sizes[11] =
	{
		{4096, 1048576, 10,0x000},
		{2048,  524288, 9, 0x800},
		{1024,  262144, 8, 0xc00},
		{ 512,  131072, 7, 0xe00},
		{ 256,   65536, 6, 0xf00},
		{ 128,   32768, 5, 0xf20},
		{  64,   16384, 4, 0xf30},
		{  32,    8192, 3, 0xf38},
		{  16,    4096, 2, 0xf3c},
		{   8,    2048, 1, 0xf3e},
		{   4,    1024, 0, 0xf3f}
	};
	int i;
	IOLog ("AGPVIA get aperture\n");
	if(agpver == 3)
	{
		UInt16 temp;
		struct aper_size_info_16 *values;
		
		values = A_SIZE_16(agp3_generic_sizes);
		temp = pciread16(hostSpace, AGP3_VIA_APSIZE);
		temp &= 0xfff;
		
		for (i = 0; i < 11; i++) 
		{
			if (temp == values[i].size_value) 
			{
				return values[i].size * 1024 * 1024;
			}
		}
	}
	else if(agpver == 2)
	{
		UInt8 temp;
		struct aper_size_info_8 *values;
		values = A_SIZE_8(via_generic_sizes);
		
		temp = pciread8(hostSpace, VIA_APSIZE);
		for (i = 0; i < 7; i++) {
			if (temp == values[i].size_value) 
			{
				return values[i].size * 1024 * 1024;
			}
		}
	}
	
	return 0;
}

//----------------------------------------------------------------------------

IOReturn AGPVIA::Configure(IOAGPDevice *master, 
						IOPhysicalAddress * address, IOPhysicalLength * length)
{
	IOReturn err;
	agpver = super::GetAGPVersion(master);
	IOLog("AGPVIA version %d", agpver);
	addrMask = 1 | ITA_COH;
	isAMD = false;
	//Check AGP version to determine config registers
	if(agpver == 3)
	{
	 	regs = via_v3_regs;
	} else if(agpver == 2)
	{
		regs = via_v2_regs;
	} else
	{
		IOLog("AGP: ERROR! AGP version not found\n");
		return -1;
	} 
 
 	IOPCIAddressSpace 	hammer = hostSpace;
	hammer.s.deviceNum = 0;
	hammer.s.functionNum = 0;
//	int i;
//	ppnum_t physPage;
// 	IOReturn err;	//Slice: I check functionNum instead of deviceNum by joblo
	// Find our hammers.
//Slice - not needed for 2.6.7
/*	n_mctrl = 0;
	for (i = 0; i < 8; i++)
	{
		UInt16 classID = (UInt16)pciread16(hammer, 0xa);
		UInt32 hamID = (UInt32)pciread32(hammer, 0);
		IOLog("AGPVIA: found hammer %08lx\n", hamID);
		if (classID == 0x0600) 
		{
			mctrl[i] = 1;
			n_mctrl++;
		} else mctrl[i] = 0;
		
		hammer.s.functionNum++;
	}
	// No additional memory controller found if n == 0
	if(!n_mctrl) n_mctrl=1;
			 
	hammer.s.deviceNum = 0;
	hammer.s.functionNum = 0;
*/
	n_mctrl=1;
	//Get aperture size
	initial_aperture = agp_via_get_aperture();
	systemLength = initial_aperture;
	if(!initial_aperture){
		IOLog("AGP: ERROR! Failed to retrieve aperture size from registers\n");
		return -1;
	}
	else if(systemLength < 0x02000000 || systemLength > 0x08000000)	
		IOLog("AGP: WARNING! Please set your aperture to 32,64 or 128MB in your BIOS for more compatibility\n");
	
		//entries = initial_aperture >> AGP_PAGE_SHIFT;
	UInt32 newApbase, tmp, curBase;	
	tmp  = pciread32(hostSpace, AGP_APBASE);
	curBase = tmp  & 0xfffffff0;
	if(*address == 0)
		newApbase = curBase;
	else newApbase = *address;
	
	if(newApbase != curBase)
	{
		UInt32 apsizeValue = pciread8(hammer, VIA_APSIZE);
		// Slice - before write we need to enable it
		pciwrite8(hammer, VIA_APSIZE, 0xff);
		pciwrite32(hammer, AGP_APBASE, 0);
		pciwrite8(hammer, VIA_APSIZE, apsizeValue);
		pciwrite32(hammer, AGP_APBASE, newApbase | (tmp & 0x0000000f));
	}
	
	
	systemBase			= newApbase;
	
		IOLog("AGPVIA: aperture [%08lx, %08lx]\n", systemBase, systemLength);

		*address = systemBase;
		*length = systemLength;
		err = kIOReturnSuccess;
	
	return ( err );
}
//----------------------------------------------------------------------------

IOReturn 
AGPVIA::InstallGATT(IOPhysicalAddress gattPhys)
{
//	UInt32 temp;
	IOReturn err;
 
 	IOPCIAddressSpace 	hammer = hostSpace;
	hammer.s.deviceNum = 0;
	hammer.s.functionNum = 0;

		// Install the gatt in each hammer and enable aperture. 
//Slice - omly 1 hammer in 2.6.7
//	for (int i = 0; i < 8; i++) 
//		{
//			if( mctrl[i])
//			{
				if (regs == via_v2_regs)
				{
					// Install the gatt. 
					pciwrite32(hammer,  regs[REG_ATTBASE], (gattPhys << 12) | 3);
					// Enable the aperture. 
					pciwrite32(hammer, regs[REG_GARTCTRL], 0x0f);

				} else
				{
					UInt32 gartctrl;
					// Install the gatt. 
					pciwrite32(hammer, regs[REG_ATTBASE], (gattPhys << 12));
					// Enable the aperture. 
					gartctrl = pciread32(hammer,  regs[REG_GARTCTRL]); //REG_ATTBASE -- Slice ??
					pciwrite32(hammer,  regs[REG_GARTCTRL], gartctrl | (3 << 7));
				}
//			hammer.s.functionNum++;
//			}
//		}
		IOLog("AGPVIA trace PCI space\n");
		for(int i=0; i<0xff; i+=16)
			{
				for(int j=0; j<15; j+=4)
					IOLog("(%02lx)=%08lx   ",(UInt32)(i+j), (UInt32)pciread32(hammer,i+j));  //for trace
				IOLog("\n");
			} 

				
		readyForPM = true;
		err = kIOReturnSuccess;
	
	return ( err );
}

//----------------------------------------------------------------------------

const char* AGPVIA::Match(UInt16 dev)
{
// AGPBSpace!
	switch (dev) {	
		case 0x0198:
			return ("VIA 8763 (P4X600) host to PCI bridge");
		case 0x0259:
			return ("VIA PM800/PN800/PM880/PN880 host to PCI bridge");
		case 0x0269:
			return ("VIA KT880 host to PCI bridge");
		case 0x0296:
			return ("VIA 3296 (P4M800) host to PCI bridge");
		case 0x0305:
			return ("VIA 82C8363 (Apollo KT133x/KM133) host to PCI bridge");
		case 0x0391:
			return ("VIA 8371 (Apollo KX133) host to PCI bridge");
		case 0x0501:
			return ("VIA 8501 (Apollo MVP4) host to PCI bridge");
		case 0x0597:
			return ("VIA 82C597 (Apollo VP3) host to PCI bridge");
		case 0x0598:
			return ("VIA 82C598 (Apollo MVP3) host to PCI bridge");
		case 0x0601:
			return ("VIA 8601 (Apollo ProMedia/PLE133Ta) host to PCI bridge");
		case 0x0605:
			return ("VIA 82C694X (Apollo Pro 133A) host to PCI bridge");
		case 0x0691:
			return ("VIA 82C691 (Apollo Pro) host to PCI bridge");
		case 0x3091:
			return ("VIA 8633 (Pro 266) host to PCI bridge");
		case 0x3099:
			return ("VIA 8367 (KT266/KY266x/KT333) host to PCI bridge");
		case 0x3101:
			return ("VIA 8653 (Pro266T) host to PCI bridge");
		case 0x3112:
			return ("VIA 8361 (KLE133) host to PCI bridge");
		case 0x3116:
			return ("VIA XM266 (PM266/KM266) host to PCI bridge");
		case 0x3123:
			return ("VIA 862x (CLE266) host to PCI bridge");
		case 0x3128:
			return ("VIA 8753 (P4X266) host to PCI bridge");
		case 0x3148:
			return ("VIA 8703 (P4M266x/P4N266) host to PCI bridge");
		case 0x3156:
			return ("VIA XN266 (Apollo Pro266) host to PCI bridge");
		case 0x3168:
			return ("VIA 8754 (PT800) host to PCI bridge");
        case 0x3188:
             return ("VIA 8385 host to PCI bridge");
		case 0x3189:
			return ("VIA 8377 (Apollo KT400/KT400A/KT600) host to PCI bridge");
		case 0x3205:
			return ("VIA 8235/8237 (Apollo KM400/KM400A) host to PCI bridge");
		case 0x3208:
			return ("VIA 8783 (PT890) host to PCI bridge");
		case 0x3258:
			return ("VIA PT880 host to PCI bridge");
	//	case 0xb188:
	//		return ("VIA K8T800Pro host to PCI bridge");
		case 0xb198:
			return ("VIA VT83xx/VT87xx/KTxxx/Px8xx host to PCI bridge");
	};
 
	return NULL;
}

//----------------------------------------------------------------------------

void AGPVIA::Flush0TLB()
{	
	UInt32 gartctrl;
	IOPCIAddressSpace 	hammer = hostSpace;
	hammer.s.deviceNum		= 0;
	hammer.s.functionNum = 0;
	// flush in each hammer
//Slice -- don't need in 2.6.7	
//	for (int i = 0; i < 8; i++) 
//	{
//		if(mctrl[i])
//		{
			if (regs == via_v2_regs)
			{
				pciwrite32(hammer, regs[REG_GARTCTRL], 0x8f);
//				pciwrite32(hammer, regs[REG_GARTCTRL], 0x0f);
			} else {
				gartctrl = pciread32(hammer, regs[REG_GARTCTRL]);
				pciwrite32(hammer, regs[REG_GARTCTRL], gartctrl & ~(1 << 7));
//				pciwrite32(hammer, regs[REG_GARTCTRL], gartctrl | (1 << 7));
			}
//		}
//		hammer.s.functionNum++;
//	}

}
//----------------------------------------------------------------------------

void AGPVIA::Flush1TLB()
{	
	UInt32 gartctrl;
	IOPCIAddressSpace 	hammer = hostSpace;
	hammer.s.deviceNum		= 0;
	hammer.s.functionNum = 0;
	//Slice -- don't need in 2.6.7	
	// flush in each hammer
//	for (int i = 0; i < 8; i++) 
//	{
//		if(mctrl[i])
//		{
			if (regs == via_v2_regs)
			{
	//			pciwrite32(hammer, regs[REG_GARTCTRL], 0x8f);
				pciwrite32(hammer, regs[REG_GARTCTRL], 0x0f);
			} else {
				gartctrl = pciread32(hammer, regs[REG_GARTCTRL]);
//				pciwrite32(hammer, regs[REG_GARTCTRL], gartctrl & ~(1 << 7));
				pciwrite32(hammer, regs[REG_GARTCTRL], gartctrl | (1 << 7));
			}
//		}
//		hammer.s.functionNum++;
//	}

}
//----------------------------------------------------------------------------

void AGPVIA::CheckTLB()
{
}



//----------------------------------------------------------------------------
//Should set the right things to do on exit here, we don't do much yet
void AGPVIA::Disable( IOAGPDevice * master )
{
	super::Disable(master);
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
