 /*-
 * Copyright (c) 2006 Jung-uk Kim <jkim@FreeBSD.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <IOKit/IOLib.h>
#include <IOKit/platform/ApplePlatformExpert.h>
#include <IOKit/pci/IOPCIBridge.h>
#include <IOKit/IOMapper.h>
#include <IOKit/IORangeAllocator.h>

#include "IOAGPBridge.h"
#include "pcireg.h"


/* GART cache control register bits. */
#define INVGART		(1<<0)
#define GARTPTEERR	(1<<1)
#define GPTE_VALID      1
#define GPTE_COHERENT   2

#define AMD64_GARTCACHECTL      0x9c
#define AMD64_MAX_MCTRL	4
/********************************************************************/
/*						AMD64 DRIVER FUNCTIONS						*/
/*							by joblo10								*/
/*						  December 2, 2006							*/	
/********************************************************************/

#define super IOAGPBridge


//----------------------------------------------------------------------------

AGPAMD64::AGPAMD64(IOService *provider)
{
	//Accessor for configRead and configWrite 
	bridge = (IOPCIDevice *)provider;
	
	//PCI Space of the host bridge
	AGPBSpace = bridge->space;
	hostSpace = AGPBSpace;
	hostSpace.s.deviceNum = 0;	
}

//----------------------------------------------------------------------------

void 
AGPAMD64::agp_amd64_apbase_fixup()
{
	IOPCIAddressSpace 	hammer = hostSpace;
	hammer.bits				= 0;
	
	hammer.s.deviceNum		= 24;
	hammer.s.functionNum = 3;
	int i;

	apbase = pciread32(hammer, AGP_AMD64_APBASE);
	IOLog("AGPAMD apbase=%08lx\n", apbase << 25);
	for (i = 0; i < n_mctrl; i++)
	{
			pciwrite32(hammer, AGP_AMD64_APBASE, apbase & ~(AGP_AMD64_APBASE_MASK & ~(UInt32)0x7f));
		hammer.s.deviceNum++;
	}
	
	hammer.s.deviceNum = 24;
	
	apbase		= apbase << 25;
	systemBase	= apbase;
}

//----------------------------------------------------------------------------

int 
AGPAMD64::agp_amd64_via_match()
{
	// XXX Some VIA bridge requires secondary AGP bridge at 0:1:0. 
	if (pciread8(AGPBSpace, PCIR_CLASS) != PCIC_BRIDGE ||
	    pciread8(AGPBSpace, PCIR_SUBCLASS) != PCIS_BRIDGE_PCI ||
	    pciread16(AGPBSpace, PCIR_VENDOR) != 0x1106 ||
	    pciread16(AGPBSpace, PCIR_DEVICE) != 0xb188 ||
	    (pciread8(AGPBSpace, AGP_VIA_AGPSEL) & 2))
		return 0;

	return 1;
}

//----------------------------------------------------------------------------
//Slice:  nonsense
void 
AGPAMD64::agp_amd64_via_init()
{
	agp_amd64_apbase_fixup();
	
}

//----------------------------------------------------------------------------

int 
AGPAMD64::agp_amd64_via_set_aperture(UInt32 aperture)
{
	UInt32 apsize;

	apsize = ((aperture - 1) >> 20) ^ 0xff;
	if ((((apsize ^ 0xff) << 20) | ((1 << 20) - 1)) + 1 != aperture)
		return -1;
	pciwrite8(hostSpace, AGP3_VIA_APSIZE, apsize);

	return 0;
}
/*
//----------------------------------------------------------------------------
//Slice: not in use
int 
AGPAMD64::agp_amd64_nvidia_match(UInt16 devid)
{
	// XXX nForce3 requires secondary AGP bridge at 0:11:0. 
	if (pciread8(AGPBSpace, PCIR_CLASS) != PCIC_BRIDGE ||
		pciread8(AGPBSpace, PCIR_SUBCLASS) != PCIS_BRIDGE_PCI ||
		pciread16(AGPBSpace, PCIR_VENDOR) != 0x10de ||
		pciread16(AGPBSpace, PCIR_DEVICE) != devid)
		return -1;
 
	return 0;
}
*/
//----------------------------------------------------------------------------
 
const char* 
AGPAMD64::Match(UInt16 dev)
{
// Slice: dev from AGPBSpace
         switch (dev) {
         case 0x7454:
                 return ("AMD 8151 AGP graphics tunnel");
         case 0x0755:
                 return ("SiS 755 host to AGP bridge");
         case 0x0760:
                 return ("SiS 760 host to AGP bridge");
         case 0x1689:
		 case 0x5246:
				//IOLog("Matched with %08lx\n",dev);
                 return ("ULi M1689 AGP Controller");
		 case 0x00d2:
                 return ("NVIDIA nForce3 AGP Controller");
         case 0x00e2:
                 return ("NVIDIA nForce3-250 AGP Controller");
         case 0x0204: //Slice - hostSpace
                 return ("VIA 8380 host to PCI bridge");
         case 0x0238: //Slice - hostSpace
                 return ("VIA 3238 host to PCI bridge");
    //     case 0x0282:  //Slice: This is a joblo's mistake. The deviceID of hostSpace while we are
						// at AGPBSpace
		 case 0xb188:
                 return ("VIA  AGP bridge for AMD"); 
				 
	//	 case 0x3188: //Slice - hostSpace
   //              return ("VIA 8385 host to PCI bridge");
 /*		 case 0x5a3f:
                 return ("ATI RS400 AGP bridge for AMD"); 
 		 case 0x7932:
                 return ("ATI RS500 AGP bridge for AMD"); 
*/		
  
         };
 
         return NULL;
 }

//----------------------------------------------------------------------------

UInt32 
AGPAMD64::agp_amd64_get_aperture(IOPCIAddressSpace dev)
{
	UInt32 i;
	UInt32 agp_amd64_table[] = {
		0x02000000,	/*   32 MB */
		0x04000000,	/*   64 MB */
		0x08000000,	/*  128 MB */
		0x10000000,	/*  256 MB */
		0x20000000,	/*  512 MB */
		0x40000000,	/* 1024 MB */
		0x80000000,	/* 2048 MB */
	};
	IOPCIAddressSpace 	hammer = hostSpace;
	hammer.s.deviceNum		= 24;
	hammer.s.functionNum	= 3;


	i = (pciread32(hammer, AGP_AMD64_APCTRL) &
		AGP_AMD64_APCTRL_SIZE_MASK) >> 1;

	if (i >= (sizeof(agp_amd64_table) / sizeof(agp_amd64_table[0])))
		return 0;

	return (agp_amd64_table[i]);
}

//----------------------------------------------------------------------------

void 
AGPAMD64::agp_amd64_uli_init()
{

	agp_amd64_apbase_fixup();
	pciwrite32(hostSpace, AGP_AMD64_ULI_APBASE,
	    (pciread32(hostSpace, AGP_AMD64_ULI_APBASE) & 0x0000000f) | apbase);
	pciwrite32(hostSpace, AGP_AMD64_ULI_HTT_FEATURE, apbase);
}

//----------------------------------------------------------------------------

int 
AGPAMD64::agp_amd64_uli_set_aperture(UInt32 aperture)
{

	switch (aperture) {
	case 0x02000000:	/*  32 MB */
	case 0x04000000:	/*  64 MB */
	case 0x08000000:	/* 128 MB */
	case 0x10000000:	/* 256 MB */
		break;
	default:
		return -1;
	}

	pciwrite32(hostSpace, AGP_AMD64_ULI_ENU_SCR,
			apbase + aperture - 1);

	return 0;
	
}

//----------------------------------------------------------------------------

void 
AGPAMD64::agp_amd64_nvidia_init()
{
	agp_amd64_apbase_fixup();
	
	pciwrite32(hostSpace, AGP_AMD64_NVIDIA_0_APBASE,
	    (pciread32(hostSpace, AGP_AMD64_NVIDIA_0_APBASE) & 0x0000000f) |
	    apbase);
		
	pciwrite32(AGPBSpace, AGP_AMD64_NVIDIA_1_APBASE1, apbase);
	pciwrite32(AGPBSpace, AGP_AMD64_NVIDIA_1_APBASE2, apbase);
}

//----------------------------------------------------------------------------

int 
AGPAMD64::agp_amd64_nvidia_set_aperture(UInt32 aperture)
{
	UInt32 apsize;

	switch (aperture) {
	case 0x02000000:	apsize = 0x0f;	break;	/*  32 MB */
	case 0x04000000:	apsize = 0x0e;	break;	/*  64 MB */
	case 0x08000000:	apsize = 0x0c;	break;	/* 128 MB */
	case 0x10000000:	apsize = 0x08;	break;	/* 256 MB */
	case 0x20000000:	apsize = 0x00;	break;	/* 512 MB */
	default:
		return -1;
	}

	pciwrite32(AGPBSpace, AGP_AMD64_NVIDIA_1_APSIZE,
	    (pciread32(AGPBSpace, AGP_AMD64_NVIDIA_1_APSIZE) & 0xfffffff0) | apsize); //Slice: no sense - bits 0-3 RO
			
	pciwrite32(AGPBSpace, AGP_AMD64_NVIDIA_1_APLIMIT1, apbase + aperture - 1);
		
	pciwrite32(AGPBSpace, AGP_AMD64_NVIDIA_1_APLIMIT2, apbase + aperture - 1);

	return 0;
}
//----------------------------------------------------------------------------

void 
AGPAMD64::agp_amd64_ATI_init()
{
	agp_amd64_apbase_fixup();
	
	pciwrite32(hostSpace, kiAPBASE, (pciread32(hostSpace, kiAPBASE) & 0x0000000f) |
	    apbase);
}

//----------------------------------------------------------------------------

int 
AGPAMD64::agp_amd64_ATI_set_aperture(UInt32 aperture)
{
	UInt32 apsize; 

	switch (aperture) {
	case 0x02000000:	apsize = 0x00;	break;	/*  32 MB */
	case 0x04000000:	apsize = 0x02;	break;	/*  64 MB */
	case 0x08000000:	apsize = 0x04;	break;	/* 128 MB */
	case 0x10000000:	apsize = 0x06;	break;	/* 256 MB */
	default:
		return -1;
	}

	pciwrite32(AGPBSpace, ATI_RS300_APSIZE, apsize | 1);
			
	return 0;
}


//----------------------------------------------------------------------------

IOReturn 
AGPAMD64::Configure(IOAGPDevice *master, 
						IOPhysicalAddress * address, IOPhysicalLength * length)
{
	IOPCIAddressSpace 	hammer = hostSpace;
	hammer.s.functionNum = 3;
	hammer.s.deviceNum = 24;
	
	int i, n;
//	ppnum_t physPage;
 	IOReturn err;	
	addrMask = 1 | ITA_COH; 
	isAMD = true;
	// Find our hammers.
		for (i = 0, n = 0; i < PCI_SLOTMAX && n < AMD64_MAX_MCTRL; i++)
		if (pciread32(hammer, 0) == 0x11031022) {
			mctrl[n] = i;
			hammer.s.deviceNum++;
			n++;
		}
 n_mctrl = n;
	if (n == 0)
	// No AMD memory controller found, exit here.
	{
		IOLog("AGPAMD: hammers not found\n");
		return -1;
	} else
		IOLog("AGPAMD: %d hammers found\n", n);
	
	hammer.s.deviceNum = 24;
	hammer.s.functionNum = 3;
	
	// Get the aperture size set in BIOS and compute the size 
	// of the GART table. 
	initial_aperture = agp_amd64_get_aperture(hostSpace);
	systemLength = initial_aperture;
	if( !systemLength)
	{
		IOLog("AGP: ERROR! Could not find aperture size from registers\n");
		return -1;
	}
	else if(systemLength < 0x02000000 || systemLength > 0x08000000)	
		IOLog("AGP: WARNING! Please set your aperture to 64 or 128MB in your BIOS for more compatibility\n");

				
	switch (pciread16(hostSpace,0)) 
		{
			case 0x10b9:	/* ULi */
				agp_amd64_uli_init();
				if (agp_amd64_uli_set_aperture(initial_aperture))
					return -1;
				break;
			case 0x10de:	/* nVidia */
				agp_amd64_nvidia_init();
				if (agp_amd64_nvidia_set_aperture(initial_aperture))
				{
					IOLog("AGP AMD nVidia: aperture not set\n");
					return -1;
					}
				break;
			case 0x1002:	/* ATI */
				agp_amd64_ATI_init();
				if (agp_amd64_ATI_set_aperture(initial_aperture))
				{
					IOLog("AGP AMD ATI: aperture not set\n");
					return -1;
					}
				break;

			case 0x1106:	/* VIA */
				int via_agp = agp_amd64_via_match();
				if (via_agp != 0) 
				{
					agp_amd64_via_init();
					if (agp_amd64_via_set_aperture(initial_aperture))
					{
						IOLog("AGP AMD VIA: aperture not set\n");
						return -1;
						}
				}
				break;
		}
	pciwrite32(hostSpace, AGP_APBASE, systemBase);
   		// Tell the caller.
		*address = systemBase;
		*length = systemLength;
		err = kIOReturnSuccess;
		IOLog("AGPAMD: aperture [%08lx, %08lx]\n", systemBase, systemLength);

 	
	return ( err );
}
//----------------------------------------------------------------------------

IOReturn 
AGPAMD64::InstallGATT(IOPhysicalAddress gattPhys)
{
	 	IOReturn err;	
		IOPCIAddressSpace 	hammer = hostSpace;	
		IOPCIAddressSpace 	via = hostSpace;	
			via.s.deviceNum = 0;  //Slice - ???
			via.s.functionNum = 0;
	
		if ((pciread16(hostSpace,0) == 0x1106) && (pciread16(via, 0x34) != 0))
		{
			pciwrite32(via, AGP3_VIA_ATTBASE, gattPhys << 12);
			pciwrite32(via, AGP3_VIA_GARTCTRL,
				pciread32(via, AGP3_VIA_GARTCTRL) | 0x180);  //Slice: it was ATTBASE
			IOLog("AGPVIA gart installed %08lx\n", pciread32(via, AGP3_VIA_ATTBASE));
		}
		Flush0TLB();
		// Install the gatt in each AMD hammer and enable aperture. 
		hammer.s.functionNum = 3;
		hammer.s.deviceNum = 24;
		
		for (int i = 0; i < n_mctrl; i++) 
		{
		
			pciwrite32(hammer, AGP_AMD64_ATTBASE, (gattPhys << 4) & AGP_AMD64_ATTBASE_MASK);
			
			pciwrite32(hammer, AGP_AMD64_APCTRL,(pciread32(hammer, AGP_AMD64_APCTRL) |
				AGP_AMD64_APCTRL_GARTEN) & ~(AGP_AMD64_APCTRL_DISGARTCPU | AGP_AMD64_APCTRL_DISGARTIO));
			pciwrite32(hammer, 0xe4,pciread32(hammer, 0xe4) | 0x40000);
			pciwrite32(hammer, 0x5c,pciread32(hammer, 0x5c) & ~0x800);
			
			hammer.s.deviceNum++;
		}
		
		hammer.s.deviceNum = 24;
		hammer.s.functionNum = 3;

		IOLog("AGPAMD trace hammer PCI space\n");
		for(int i=0; i<0xff; i+=16)
			{
				for(int j=0; j<15; j+=4)
					IOLog("(%02lx)=%08lx   ",(UInt32)(i+j), (UInt32)pciread32(hammer,i+j));  //for trace
				IOLog("\n");
			} 

		IOLog("AGPAMD trace host PCI space\n");
		for(int i=0; i<0x2f; i+=16)
		{
			for(int j=0; j<15; j+=4)
				IOLog("(%02lx)=%08lx   ",(UInt32)(i+j), (UInt32)pciread32(via,i+j));  //for trace
			IOLog("\n");
		} 
		
		Flush1TLB();
		
		// For PM. It tells our driver can now sleep and wake-up.
		readyForPM = true;
		err = kIOReturnSuccess;

 	
	return ( err );
}

//----------------------------------------------------------------------------

void 
AGPAMD64::Flush0TLB()
{
	IOPCIAddressSpace 	hammer = hostSpace;
	hammer.s.functionNum = 3;
	hammer.s.deviceNum		= 24;
	
	UInt32 tmp;
		
		for (int i = 0; i < n_mctrl; i++) 
		{
		
				tmp = pciread32 (hammer, AMD64_GARTCACHECTL);
				tmp &= ~INVGART;
				pciwrite32(hammer, AMD64_GARTCACHECTL, tmp);
		
			hammer.s.deviceNum++;
		}
}
//----------------------------------------------------------------------------

void 
AGPAMD64::Flush1TLB()
{
	IOPCIAddressSpace 	hammer = hostSpace;
	hammer.s.functionNum = 3;
	hammer.s.deviceNum		= 24;
	UInt32 tmp;
		
		for (int i = 0; i < n_mctrl; i++) 
		{
		
				tmp = pciread32 (hammer, AMD64_GARTCACHECTL);
				tmp |= INVGART;
				pciwrite32(hammer, AMD64_GARTCACHECTL, tmp);
		
			hammer.s.deviceNum++;
		}
	
	hammer.s.deviceNum = 24;
	hammer.s.functionNum = 3;
	UInt32 w;
	/* Make sure the hardware actually executed the flush. */
    do { 
		w = pciread32(hammer, AMD64_GARTCACHECTL);
	} while (w & 1);
}
//----------------------------------------------------------------------------

void AGPAMD64::CheckTLB()
{
}



//----------------------------------------------------------------------------
/*
int 
AGPAMD64::BindMemory(IOMemoryDescriptor * mem, IOByteCount agpOffset)
{
//    UInt32		offset = 0;
	//UInt32	mem_phys;
	IOByteCount	len; //, len2;
	IOPhysicalAddress	phys, phys2;
// 	UInt32 temp;
//	temp = pciread32(hostSpace, AGP_AMD64_APCTRL);
//	pciwrite32(hostSpace, AGP_AMD64_APCTRL, temp & ~AGP_AMD64_APCTRL_GARTEN);		
	IOByteCount         agpFlushStart, flushLength;
	phys = mem->getPhysicalAddress();
	len = mem->getLength();
	UInt32 gartOffset = agpOffset >> 10;
	agpFlushStart = gartOffset;
	//agpOffset >> 10;	//	/= (page_size / 4);
		
//	while( (phys = mem->getPhysicalSegment(offset, &len)))
//	{
		phys += agpOffset;
	//	offset += len;
	//	len = (len + PAGE_MASK) & ~PAGE_MASK;
		while( len > 0) 
		{
			OSWriteLittleInt32(gartArray, gartOffset, phys); // | addrMask);
			gartOffset += 4;
			phys += page_size;
			len -= page_size;
		}
//	}
	phys2 = OSReadLittleInt32(gartArray, gartOffset-4);
	flushLength = (gartOffset - agpFlushStart);
	flush_dcache( ((vm_offset_t)gartArray)+agpFlushStart, flushLength, false);

	agp_flush_cache();
	FlushTLB();
//	temp = pciread32(hostSpace, AGP_AMD64_APCTRL);
//	pciwrite32(hostSpace, AGP_AMD64_APCTRL, temp | AGP_AMD64_APCTRL_GARTEN);		
	

	return( 0 );
}
*/
//----------------------------------------------------------------------------

void 
AGPAMD64::Disable( IOAGPDevice * master )
{						
	super::Disable(master);
}

//----------------------------------------------------------------------------
